@@ -1,4 +1,4 @@
-from numpy import array, matmul, pad, delete, sqrt , mean, tanh, add
+from numpy import array, matmul, pad, delete, sqrt , mean, tanh, add, arange
 from numpy import maximum as mx
 from numpy.random import randn , seed
 from scipy.special import expit
@@ -33,39 +33,40 @@ class model:
                 2 neurons, two hidden layers with 3 neurons each(without 
                 counting bias) and an output layer of 1 neuron""")
             return -1
-        self.architecture=architecture
+        self.architecture=architecture[:]
+        architecture.append(0)
         if mtype=="relu":
             self.W=array([randn(j,i+1)*sqrt(2/(i+1)) for i,j in zip(architecture[0::1],
-                architecture[1::1])])
+                architecture[1::1])], dtype=object)[arange(len(architecture)-2)]
             self.currentmodeltype="relu"
         elif mtype == "sigmoid":   
             self.W=array([randn(j,i+1)*sqrt(1/(i+1)) for i,j in zip(architecture[0::1],
-                architecture[1::1])])
+                architecture[1::1])], dtype=object)[arange(len(architecture)-2)]
             self.currentmodeltype="sigmoid"
         elif mtype == "tanh":   
             self.W=array([randn(j,i+1)*sqrt(1/(i+1)) for i,j in zip(architecture[0::1],
-                architecture[1::1])])
+                architecture[1::1])], dtype=object)[arange(len(architecture)-2)]
             self.currentmodeltype="tanh"
         else:
             print("model type is unknown or not set properly")
             return -1
-        print(f"\nModel initialized \n\n \t Architecture = {architecture} \n \t Model type = {self.currentmodeltype}\n\n Weights :")
+        print(f"\nModel initialized \n\n \t Architecture = {self.architecture} \n \t Model type = {self.currentmodeltype}\n\n Weights :")
         for i in range(len(self.W)):
             self.W[i][:,0]=0#Initializes bias weights to 0
             print('W[%d]=\n'%i,self.W[i],'\n')
         return None
     def weights(self,plot=False):
-        '''prints out the weight matrixes w1 and w2'''
+        """prints out the weight matrixes w1 and w2"""
+        for i in range(len(self.W)):
+            print('W[%d]=\n'%i,self.W[i],'\n')
         if plot:
             nnplotter.plotinit()
             for i in range(len(self.W)):
                 nnplotter.plotweights(self.W[i],i)
             nnplotter.plt.show()
-        for i in range(len(self.W)):
-            print('W[%d]=\n'%i,self.W[i],'\n')
         return 0
     def predictrelu(self,x):
-        '''returns the network output of a specific input specifically using relu activation'''
+        """returns the network output of a specific input specifically using relu activation"""
         l=len(self.W)-1
         q=lambda z,y:mx(matmul(pad(x,((0,0),(1,0)),
           'constant',constant_values=1),self.W[z].T),0)if z==0 else (expit(matmul(pad(q(z-1,y),((0,0),(1,0)),
@@ -73,27 +74,27 @@ class model:
             'constant',constant_values=1),self.W[z].T),0))
         return q(l,l)
     def predictsigmoid(self,x):
-        '''returns the network output of a specific input specifically using sigmoid activation'''
+        """returns the network output of a specific input specifically using sigmoid activation"""
         p=lambda z:expit(matmul(pad(x,((0,0),(1,0)),
             'constant',constant_values=1),self.W[z].T))if z==0 else expit(matmul(
                 pad(p(z-1),((0,0),(1,0)),
                        'constant',constant_values=1),self.W[z].T))
         return p(len(self.W)-1)
     def predicttanh(self,x):
-        '''returns the network output of a specific input specifically using tanh activation'''
+        """returns the network output of a specific input specifically using tanh activation"""
         p=lambda z:tanh(matmul(pad(x,((0,0),(1,0)),
             'constant',constant_values=1),self.W[z].T))if z==0 else tanh(matmul(
                 pad(p(z-1),((0,0),(1,0)),
                        'constant',constant_values=1),self.W[z].T))
         return p(len(self.W)-1)
     def predict(self,x ):
-        '''returns the network output of a specific input specifically
+        """returns the network output of a specific input specifically
             NOTE:
                 if a single input is given to predict funcion it must be of shape
                 (1,n) {n = no of input neurons}  
                 Example: [1 , 1] has a shape of (2,) which is not accepted yet
                 but [[1, 1]] has a shape of (1,2) which is desired if single input
-                You know what you are doing :) '''
+                You know what you are doing :) """
         if self.currentmodeltype == "relu":
             return(self.predictrelu(x))
         elif self.currentmodeltype == "sigmoid":
@@ -112,8 +113,8 @@ class model:
         """Loads a model saved using aimaster modules"""
         with open(f"{filename}",'rb') as file:
             return load(file)
-    def train(self,x,y,iterations,learningrate,plot=False,printy=True,printw=True,vmode="queue"):
-        '''activation argument is used to select activation for neural network
+    def train(self,x,y,iterations,learningrate,plot=False,printy=True,printw=True,vmode="queue",boost=0,L2=0):
+        """activation argument is used to select activation for neural network
         Specify Activation argument as < activation="sigmoid" > or "relu"
         Over the iterations, this function optimizes the values of all weights
         to reduce output error.
@@ -126,18 +127,18 @@ class model:
         less than required (result: slow training). So feel free to experiment with 
         different values as this module is for basic understanding and experiments :)
         Adaptive learning rate will be introduced in future.
-        '''
+        """
         if self.currentmodeltype=="sigmoid":
-            self.trainsigmoid(x,y,iterations,learningrate,plot,printy,printw,vmode)
+            self.trainsigmoid(x,y,iterations,learningrate,plot,printy,printw,vmode,boost,L2)
         elif self.currentmodeltype=="relu":
-            self.trainrelu(x,y,iterations,learningrate,plot,printy,printw,vmode)
+            self.trainrelu(x,y,iterations,learningrate,plot,printy,printw,vmode,boost,L2)
         elif self.currentmodeltype=="tanh":
-            self.traintanh(x,y,iterations,learningrate,plot,printy,printw,vmode)
+            self.traintanh(x,y,iterations,learningrate,plot,printy,printw,vmode,boost,L2)
         else:
-            print("Either currentmodeltype not set or corrupt.check again")
+            print("Either currentmodeltype not set or corrupt. Check again")
             return None
-    def trainsigmoid(self,x,y,iterations,learningrate,plot=False,printy=True,printw=True,vmode="queue"):
-        '''Uses Sigmoid Activation.'''
+    def trainsigmoid(self,x,y,iterations,learningrate,plot=False,printy=True,printw=True,vmode="queue",boost=0,L2=0):
+        """Uses Sigmoid Activation."""
         if plot:
             if vmode=="queue":
                 event_q = Queue()
@@ -154,8 +155,11 @@ class model:
                 print("visualization mode unknown. Turning off plotting")
                 plot=False
         Wcorr=self.W*0
-        tmp=0
         lw= len(self.W)
+        lx=len(x)
+        if boost:
+            tmp=Wcorr
+            boostcounter = abs(y-self.predict(x)).argmax()
         result=[[] for i in range(lw)]
         #Lsum=[[] for i in range(len(W))]
         p=lambda z:expit(matmul(pad(x,((0,0),(1,0)),
@@ -165,30 +169,34 @@ class model:
             for k in range(iterations):
                 for i in range(lw-1,-1,-1):
                     result[i]=pad(p(i),((0,0),(1,0)),'constant',constant_values=1)
-                tmp=0
                 for i in range(len(x)):
+                    if boost:
+                        boostcounter = abs(y-self.predict(x)).argmax()
                     X=pad(x[i],((1,0)),'constant',constant_values=1)
                     for j in range(lw-1,-1,-1):
                         if j==lw-1:
                             Wcorr[j]=array([(result[j][i]-y[i])*(result[j][i]*(1-result[j][i]))])#(pred - expected)*(derivative of activation)
                         else:
                             Wcorr[j]=(matmul(Wcorr[j+1][0][1:],self.W[j+1])*array([(result[j][i]*(1-result[j][i]))]))
-                    #print ( "Wcorr", Wcorr)
-                    tmp=tmp+Wcorr
-                    #print("tmp",tmp)
-                #Wcorr=tmp
-                #tmp=0
+                    if boost:
+                        tmp=tmp+Wcorr
+                        if i == boostcounter:
+                            tmp=Wcorr
+                        if(any([abs(subw.max()) > boost for subw in tmp])):
+                            tmp = Wcorr
                     for j in range(lw-1,-1,-1):
                         if j==0:
-                            self.W[0]=self.W[0]-learningrate*delete(matmul(Wcorr[0].T,array([X])),0,0)
+                            self.W[0]=self.W[0]-learningrate*(delete(matmul(Wcorr[0].T,array([X])),0,0)+((L2/2*lx)*self.W[0]))
                         else:
-                            self.W[j]=self.W[j]-learningrate*delete(matmul(Wcorr[j].T,array([result[j-1][i]])),0,0)
-                    for j in range(lw-1,-1,-1):
-                        if j==0:
-                            self.W[0]=self.W[0]-learningrate*delete(matmul(tmp[0].T,array([X])),0,0)
-                        else:
-                            self.W[j]=self.W[j]-learningrate*delete(matmul(tmp[j].T,array([result[j-1][i]])),0,0)
-                Loss = (mean((self.predictsigmoid(x)-y)**2))/len(x)
+                            self.W[j]=self.W[j]-learningrate*(delete(matmul(Wcorr[j].T,array([result[j-1][i]])),0,0)+((L2/2*lx)*self.W[j]))
+                    #following loop lines currently boost the sigmoid(experimental).
+                    if boost :
+                        for j in range(lw-1,-1,-1):
+                            if j==0:
+                                self.W[0]=self.W[0]-learningrate*(delete(matmul(tmp[0].T,array([X])),0,0)+((L2/2*lx)*self.W[0]))
+                            else:
+                                self.W[j]=self.W[j]-learningrate*(delete(matmul(tmp[j].T,array([result[j-1][i]])),0,0)+((L2/2*lx)*self.W[j]))
+                Loss = (mean((self.predictsigmoid(x)-y)**2))
                 if plot:
                     if vmode == "queue":
                         try:
@@ -201,8 +209,9 @@ class model:
                             pconn.send(self.W)
                             pconn3.send([k,Loss])
                 if printy:
-                    print(self.predictsigmoid(x))
-                print('iteration : {}'.format(k+1))
+                    print(str(self.predictsigmoid(x))+'\n iteration :'+ str(k+1))
+                else:
+                    print('iteration : {}'.format(k+1))
         except KeyboardInterrupt:
             pass
         if printw:
@@ -212,8 +221,8 @@ class model:
             event_q.put("close")
             q.join()
         return 0
-    def traintanh(self,x,y,iterations,learningrate,plot=False,printy=True,printw=True,vmode="queue"):
-        '''Uses tanh Activation.'''
+    def traintanh(self,x,y,iterations,learningrate,plot=False,printy=True,printw=True,vmode="queue",boost=0,L2=0):
+        """Uses tanh Activation."""
         if plot:
             if vmode=="queue":
                 event_q = Queue()
@@ -252,7 +261,7 @@ class model:
                             self.W[0]=self.W[0]-learningrate*delete(matmul(Wcorr[0].T,array([X])),0,0)
                         else:
                             self.W[j]=self.W[j]-learningrate*delete(matmul(Wcorr[j].T,array([result[j-1][i]])),0,0)
-                Loss = (mean((self.predicttanh(x)-y)**2))/len(x)
+                Loss = (mean((self.predicttanh(x)-y)**2))
                 if plot:
                     if vmode == "queue":
                         try:
@@ -265,8 +274,9 @@ class model:
                             pconn.send(self.W)
                             pconn3.send([k,Loss])
                 if printy:
-                    print(self.predicttanh(x))
-                print('iteration : {}'.format(k+1))
+                    print(str(self.predicttanh(x))+'\n iteration :'+ str(k+1))
+                else:
+                    print('iteration : {}'.format(k+1))
         except KeyboardInterrupt:
             pass
         if printw:
@@ -276,8 +286,8 @@ class model:
             event_q.put("close")
             q.join()
         return 0
-    def trainrelu(self,x,y,iterations,learningrate,plot=False,printy=True,printw=True,vmode="queue"):
-        '''Relu Activation for Hidden layers and Sigmoid on final output.'''
+    def trainrelu(self,x,y,iterations,learningrate,plot=False,printy=True,printw=True,vmode="queue",boost=0,L2=0):
+        """Relu Activation for Hidden layers and Sigmoid on final output."""
         if plot:
             if vmode=="queue":
                 event_q = Queue()
@@ -297,8 +307,6 @@ class model:
         lw= len(self.W)
         result=[[] for i in range(lw)]
         #Lsum=[[] for i in range(len(W))]
-        if plot:
-            nnplotter.plotinit()
         q=lambda z,y:mx(matmul(pad(x,((0,0),(1,0)),
           'constant',constant_values=1),self.W[z].T),0)if z==0 else (expit(matmul(pad(q(z-1,y),((0,0),(1,0)),
             'constant',constant_values=1),self.W[z].T)) if (z == y) else mx(matmul(pad(q(z-1,y),((0,0),(1,0)),
@@ -319,7 +327,7 @@ class model:
                             self.W[0]=self.W[0]-learningrate*delete(matmul(Wcorr[0].T,array([X])),0,0)
                         else:
                             self.W[j]=self.W[j]-learningrate*delete(matmul(Wcorr[j].T,array([result[j-1][i]])),0,0)
-                Loss = (mean((self.predictrelu(x)-y)**2))/len(x)
+                Loss = (mean((self.predictrelu(x)-y)**2))
                 if plot:
                     if vmode == "queue":
                         try:
@@ -332,8 +340,9 @@ class model:
                             pconn.send(self.W)
                             pconn3.send([k,Loss])
                 if printy:
-                    print(self.predictrelu(x))
-                print('iteration : {}'.format(k+1))
+                    print(str(self.predictrelu(x))+'\n iteration :'+ str(k+1))
+                else:
+                    print('iteration : {}'.format(k+1))
         except KeyboardInterrupt:
             pass
         if printw:
@@ -361,7 +370,7 @@ class model:
                 for i in range(len(tmp[0])):
                     nnplotter.plotweights(tmp[0][i],i)
                 nnplotter.ax.text(0,-0.25,s="iteration {: <5} Loss = {: <10}".format(tmp[1],tmp[2]))
-                nnplotter.plt.pause(0.00000001)
+                nnplotter.plt.pause(0.001)
                 send_q.put("Send")
     def processplotterpipe(self,cconn,cconn2,cconn3):
         nnplotter.plotinit()
@@ -381,7 +390,7 @@ class model:
             for i in range(len(tmp)):
                 nnplotter.plotweights(tmp[i],i)
             nnplotter.ax.text(0,-0.25,s="iteration {: <5} Loss = {: <10}".format(str(k[0]),str(k[1])))
-            nnplotter.plt.pause(0.00000001)
+            nnplotter.plt.pause(0.001)
             try:
                 cconn2.send("Send")
             except:
